/*
 * Copyright (c) 2003 Andrew Stikheev. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * $Id: scand.c,v 1.21 2004/09/07 09:44:41 sand Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <syslog.h>
#include <sys/sysctl.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <signal.h>

#include <fcntl.h>
#include <time.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netdb.h>

#include <math.h>
#include <err.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>

#include "statistics.h"
#include "base.h"
#include "mem.h"
//------------------------------------------------------------------------------
static void		DaemonMode ();
static void		InitiateShutdown (int);
static void		Shutdown (int);
static void		ParseArgs (int argc, char **argv);
static void		Usage (const char *);
static void		KillDaemon ();
static void		ReadConfigFile ( const char *fileName );
static void		ParseOption ( const char* option, const char* parms ); 
void			SigAlarm(int signo);
//------------------------------------------------------------------------------
void			*hashTable;
//------------------------------------------------------------------------------

systemParameters	*sys;
static systemVariables	*sysV;

static	int		verbose		= 0; // disabled
	int		background  	= 0; // not a daemon
static	int		running     	= 1; 
int			fdLog 		= -1;

char                    packetBuf[IP_MAXPACKET];
////////////////////////////////////////////////////////////////////////////////
int main (int argc, char** argv)
{
	int                     packetLen,write;
	struct sockaddr_in      packetAddr;

	int			divert;
	int      		addrSize; 
	struct sockaddr_in	addr;

	sys = (systemParameters *) malloc(sizeof(systemParameters));
	if (sys == NULL ) Quit("malloc");
	bzero(sys,sizeof(systemParameters));

	sys->numClusters	= 1009;
	sys->clusterSize	= 4;
	sys->blockIfAny		= 1;
	sys->enableBlock	= 1;
	sys->divertPort		= 9999;
	sys->maxWeight[0]	= 240;
	sys->maxWeight[1]	= 240;
	sys->maxWeight[2]	= 240;
	sys->blkWeight[0]	= 120;
	sys->blkWeight[1]	= 120;
	sys->blkWeight[2]	= 120;
	sys->minWeight[0]	= 0;
	sys->minWeight[1]	= 0;
	sys->minWeight[2]	= 0;
	sys->factor[0][0]	= 1;
	sys->factor[0][1]	= 0;
	sys->factor[1][0]	= 1;
	sys->factor[1][1]	= 0;
	sys->factor[2][0]	= 1;
	sys->factor[2][1]	= 1;
	sys->maxConnects[0]	= 3;
	sys->maxConnects[1]	= 3;
	sys->maxConnects[2]	= 1;

	strcpy(sys->pidFile,"/var/run/scand.pid");
/*
 * Parse options.
 */
	ParseArgs(argc, argv);
/*
 * Open syslog channel.
 */
	openlog ("scand", LOG_CONS | LOG_PID | (verbose ? LOG_PERROR : 0),
		 LOG_DAEMON);

        alarm(1);

//-----------------------------------------------------------------------
	divert = socket (PF_INET, SOCK_RAW, IPPROTO_DIVERT);
	if (divert == -1)
		Quit ("Unable to create divert socket.");
/*
 * Bind socket.
 */

	addr.sin_family		= AF_INET;
	addr.sin_addr.s_addr	= INADDR_ANY;
	addr.sin_port		= htons(sys->divertPort);

	if ( bind (divert, (struct sockaddr*) &addr, sizeof addr) == -1 )
		Quit ("Unable to bind divert socket.");


	if ( !verbose )	DaemonMode ();
/*
 * Catch signals to manage shutdown 
 */
	siginterrupt(SIGTERM, 1);
	siginterrupt(SIGHUP, 1);
	signal (SIGTERM, InitiateShutdown);

//	if (verbose) fprintf(stderr,"start pipe\n");
//------------------------------------------------------------------------------

	while (running) {

		/*
 		 * Get packet from socket.
		 */
		addrSize = sizeof packetAddr;
        	packetLen = recvfrom (	divert,
					packetBuf,
                              		sizeof packetBuf,
                              		0,
                              		(struct sockaddr*) &packetAddr,
                              		&addrSize );

        	if ( packetLen == -1 ) {

			sysV->readPacketsFailure++;
			continue;
		}
		sysV->readPackets++;

		/*
		 * Handle packet.
		 */
        	if ( HandlePacket( packetBuf ) && sys->enableBlock ) {
		
			sysV->blockPackets++;
			continue;
		}

		/*
 		 * Put packet back for processing.
 		 */
		write =	sendto (	divert, 
					packetBuf,
       					packetLen, 
					0,
					(struct sockaddr*) &packetAddr,
					addrSize );

		if ( write != packetLen )  sysV->writePacketsFailure++;
	}
//------------------------------------------------------------------------------
	if (background)
		unlink (sys->pidFile);

	if ( fdLog >= 0  ) close(fdLog);

	if ( sys->mapFile ) unlink(sys->mapFile);

	return 0;
}

////////////////////////////////////////////////////////////////////////////////
static void DaemonMode ()
{
	FILE*	pidFile;

 	daemon(0,0);
	background = 1;

	pidFile = fopen (sys->pidFile, "w");
	if (pidFile) {

		fprintf (pidFile, "%d\n", getpid ());
		fclose (pidFile);
	}
}

////////////////////////////////////////////////////////////////////////////////
static void InitiateShutdown (int sig)
{
/*
 * Start timer to allow kernel gracefully
 * shutdown existing connections when system
 * is shut down.
 */
	siginterrupt(SIGALRM, 1);
	signal (SIGALRM, Shutdown);
	alarm (10);
}

////////////////////////////////////////////////////////////////////////////////
static void Shutdown (int sig)
{
	running = 0;
}

////////////////////////////////////////////////////////////////////////////////
static void KillDaemon()
{
	FILE *pidFile;
	long p;

	pidFile = fopen( sys->pidFile, "r" );
	if ( !pidFile )
		Quit("fopen"); 
	
	if ( !fscanf( pidFile , "%ld", &p ) )
		Quit("fscanf");

	if ( kill( (pid_t) p , SIGTERM ) )
		Quit("kill");         

	printf("Signal SIGTERM to process %ld was send\n", p );
	exit(1);
}
////////////////////////////////////////////////////////////////////////////////
enum Option {

	DivertPort,
	Verbose,
	Help,
	LogFile,
	MapFile,
	BlkFile,
	PidFile,
	Kill,
	BlockingMode,
	ConfigFile,
	MaxWeight,
        BlkWeight,
	MinWeight,
	FactorInHash,
	FactorNotInHash,
	MaxConnects,
	BlockIfAny,
	ClusterSize,
	NumClusters,
};

enum Param {
        
        YesNo,
        Numeric,
	NumericList,
        String,
	None,
};

struct OptionInfo {
        
        enum Option             type;
        enum Param              parm;
        const char*             parmDescription;
        const char*             description;
        const char*             name; 
        const char*             shortName;
};
static struct OptionInfo optionTable[] = {

        { Verbose,
                None,
                "",
                "verbose mode, some useful information for debugging",
                "verbose",
                "v" },
        { Help,
                None,
                "",
                "Usage",
                "help",
                "h" },

        { DivertPort,
                Numeric,
                "Port",
                "set divert port",
                "divert_port",
                "p" },
	{ ConfigFile,
                String,
                "file_name",
                "read options from configuration file",
                "config",
                "f" },

        { BlkFile,
                String,
                "filename",
                "Save logs about blocked host to the specific file",
                "block_log",
                "B" },

	{ LogFile,
                String,
                "log_file",
                "Save TCP SYN and UDP statistics to the specific file",
                "scan_log",
                "l" },
        { MapFile,
                String,
                "map_file",
                "Use mmap(2) file to store hash table",
                "map_file",
                "m" },
        { PidFile,
                String,
                "pid_file",
                "Specify an alternate file in which to store the process ID",
                "pid_file",
                "P" },

        { Kill,
                None,
                "",
                "Send signal SIGTERM to scand",
                "quit",
                "q" },
	{ BlockingMode,
		YesNo,
		"[yes|no]",
		"Disable blocking mode",
		"locking_mode",
		"b" },
	{ MaxWeight,
		NumericList,
		"tcp udp icmp",
		"Set MaxWeight for all types",
		"max_weight",
		"W" },
        { BlkWeight,
                NumericList,
                "tcp udp icmp",
                "Set BlkWeight for all types",
                "blk_weight",
                "Y" },
        { MinWeight,
                NumericList,
                "tcp udp icmp",
                "Set MinWeight for all types",
                "max_weight",
                "w" },
        { FactorInHash,
                NumericList,
                "tcp udp icmp",
                "Set FactorInHash for all types",
                "factor_in_hash",
                "I" },
        { FactorNotInHash,
                NumericList,
                "tcp udp icmp",
                "Set FactorNotInHash for all types",
                "factor_not_in_hash",
                "i" },
        { MaxConnects,
                NumericList,
                "tcp udp icmp",
                "Set MaxConnects for all types",
                "max_connects",
                "o" },
        { BlockIfAny,
                YesNo,
                "[yes|no]",
                "Block only flood type traffic." ,
                "block_if_any",
                "a" },
        { NumClusters,
                Numeric,
                "NumClusters",
                "set number of clusters",
                "num_clusters",
                "C" },
        { ClusterSize,
                Numeric,
                "ClusterSize",
                "set number elements in cluster",
                "cluster_size",
                "c" }
};
////////////////////////////////////////////////////////////////////////////////
static void ParseOption( const char* option, const char* parms )
{
        int                     i;
        int                     j;
        struct OptionInfo*      info;
        int                     yesNoValue;
        int                     numValue;
	int			numValueList[NUM_TYPES];
        const char*             strValue;
        int                     max;
        char*                   end;
	char*			p;
/*
 * Find option from table.
 */
        max = sizeof (optionTable) / sizeof (struct OptionInfo);
        for (i = 0, info = optionTable; i < max; i++, info++) {

	if (!strcmp (info->name, option))
                        break;

                if (info->shortName)
                        if (!strcmp (info->shortName, option))
                                break;
        }

        if (i >= max) {

                warnx ("unknown option %s", option);
                Usage ("");
        }

        yesNoValue      = 0;
        numValue        = 0;
        strValue        = NULL;
        for(j = 0 ; j < NUM_TYPES ; j++ ) numValueList[j] = 0;

	switch (info->parm) {

	case None:
                if (parms)
                        errx (1, "%s does not take parameters", option);
                break;
        case YesNo:
                if (!parms)
                        parms = "yes";

                if (!strcmp (parms, "yes"))
                        yesNoValue = 1;
                else
                        if (!strcmp (parms, "no"))
                                yesNoValue = 0;
                        else
                                errx (1, "%s needs yes/no parameter", option);
                break;

        case Numeric:
                if (parms)
                        numValue = strtol (parms, &end, 10);
                else
                        end = NULL;

                if (end == parms)
                        errx (1, "%s needs numeric parameter", option);
                break;
	case NumericList:
		
		j = 0;
		i =  strtol(parms, &end, 10);
		while (parms != end ) {
			numValueList[j++] = i;
			parms = end;
			i =  strtol(parms, &end, 10);
		}

		if ( j != NUM_TYPES ) {
                        errx (1, "%s needs %d numeric parameters", 
							option,
							NUM_TYPES);
		}
		
		break;
        case String:
                strValue = parms;
                if (!strValue)
                        errx (1, "%s needs parameter", option);
                break;
	}

	switch (info->type) {

        case Verbose:
                verbose = yesNoValue;
                break;

	case DivertPort:
		sys->divertPort =  numValue; 
		break;

	case Help:
		Usage("");
		break;

	case  LogFile:

		if (strlen(strValue) >= PATH_MAX)  Usage("filename too long");
		strcpy(sys->logFile,strValue);

		if ( !strcmp(sys->logFile,"stdout") ) {

			fdLog = fileno (stdout);
			if ( fcntl(fdLog,F_SETFL,O_NONBLOCK) == -1 ) {

				Quit("fcntl");
			}
		} else if ((fdLog=open( sys->logFile, O_NONBLOCK | O_CREAT | O_WRONLY,
					S_IRUSR    | S_IWUSR | S_IRGRP    | S_IROTH )) 
					< 0 )  Quit("open");

		if ( lseek(fdLog, 0 ,SEEK_END) == -1) 
			Quit ("lseek");
		break;

	case MapFile:
		
		if (strlen(strValue) >= PATH_MAX)  Usage("filename too long");
		strcpy(sys->mapFile,strValue);
		break;
	case PidFile:

		if (strlen(strValue) >= PATH_MAX)  Usage("filename too long");
		strcpy(sys->pidFile,strValue);
		break;

        case Kill:

		KillDaemon();
		break;

        case BlockingMode:

		sys->enableBlock =  1 - yesNoValue;
		break;

        case ConfigFile:

		ReadConfigFile(strValue);
		break;

	case BlkFile:

		if (strlen(strValue) >= PATH_MAX)  Usage("filename too long");
		strcpy(sys->blkFile,strValue);
		strcpy(sys->blTFile,strValue);

		p = sys->blTFile + strlen(sys->blTFile) -1;
		*p = 'T';
                break;

	case MaxWeight:
		for(j = 0 ; j < NUM_TYPES;j++) 
			sys->maxWeight[j] = numValueList[j];
		break;

	case BlkWeight:
		for(j = 0 ; j < NUM_TYPES;j++) 
			sys->blkWeight[j] = numValueList[j];
		break;

	case MinWeight:
		for(j = 0 ; j < NUM_TYPES;j++) 
			sys->minWeight[j] = numValueList[j];
		break;

	case FactorInHash:
		for(j = 0 ; j < NUM_TYPES;j++) 
			sys->factor[j][0] = numValueList[j];
		break;

	case FactorNotInHash:
		for(j = 0 ; j < NUM_TYPES;j++) 
			sys->factor[j][1] = numValueList[j];
		break;

	case MaxConnects:
		for(j = 0 ; j < NUM_TYPES;j++) 
			sys->maxConnects[j] = numValueList[j]; 
		break;
	case BlockIfAny:
		sys->blockIfAny = 1 - yesNoValue;
		break;
	case NumClusters:
		sys->numClusters = numValue;
		break;
	case ClusterSize:
		sys->clusterSize = numValue;
		break;
	}
}
////////////////////////////////////////////////////////////////////////////////
static void ParseArgs(int argc, char ** argv)
{

	char ch[] = {0,0};
	int keys;
	int i;
	int memorySize = 0;
	int fdMmapedFile = -1;

	while ((keys = getopt(argc, argv,"ahvqbl:p:m:f:B:W:w:Y:i:I:c:C:o:")) != -1){

		ch[0] = (char) keys;
		ParseOption(ch, optarg);

	}

	for ( i = 0; i < NUM_TYPES; i++ ) 
		memorySize += sizeof(connectItem) * sys->maxConnects[i];
	
	memorySize = 	sys->numClusters * ( sizeof(cluster) + 
			sys->clusterSize * ( sizeof (statisticItem) + memorySize));

	memorySize += sizeof(systemParameters) + sizeof(systemVariables);

	printf("memorySize %d\n",memorySize);

		
	
	if (!strcmp(sys->mapFile,"") ) { 

		hashTable = malloc(memorySize);
		bzero(hashTable, memorySize);
		if (!hashTable) Quit("Not enough memory.");
	} else {
                fdMmapedFile = open(    sys->mapFile,
                                        O_RDWR     | O_CREAT |
                                        O_TRUNC,
                                        S_IRUSR    | S_IWUSR |
                                        S_IRGRP    | S_IROTH );
                if ( fdMmapedFile < 0 )
                        Quit ("Unable to open file to mmap.");

                if ( ftruncate(fdMmapedFile, memorySize) )
                        Quit("ftruncate");

                hashTable = mmap(	0, 
					memorySize,
                                       	PROT_READ | PROT_WRITE, 
					MAP_NOSYNC | MAP_SHARED,
                               		fdMmapedFile, 
					0);
                if ( errno )
                        Quit("mmap");

                if ( close(fdMmapedFile) == -1 )
                        Quit("close");

	}

	hashTable = InitizializeMemory(	hashTable,
				&sys,
				&sysV);

	 fprintf( stderr,"verbose = %d   port = %d put = %d mapFile = %s block = %d\n",
	 verbose, sys->divertPort, fdLog, sys->mapFile,sys->enableBlock ); 

}
////////////////////////////////////////////////////////////////////////////////
static void ReadConfigFile ( const char *fileName )
{
	FILE*   file;
	char    *buf;
	size_t  len;
	char    *ptr, *p;
	char*   option;

	file = fopen (fileName, "r");
	if (!file)
		Quit("Cannot open config file.");

	while ((buf = fgetln(file, &len)) != NULL) {

		if (buf[len - 1] == '\n')
			buf[len - 1] = '\0';
		else
			Quit(	"config file format error:" 
				"last line should end with newline");
		
/*
 * Check for comments, strip off trailing spaces.
 */
		if ((ptr = strchr(buf, '#')))
			*ptr = '\0';
		for (ptr = buf; isspace(*ptr); ++ptr)
			continue;
		if (*ptr == '\0')
			continue;
		for (p = strchr(buf, '\0'); isspace(*--p);)
			continue;
		*++p = '\0';

/*
 * Extract option name.
 */
		option = ptr;
		while (*ptr && !isspace (*ptr))
			++ptr;

		if (*ptr != '\0') {

			*ptr = '\0';
			++ptr;
		}
/*
 * Skip white space between name and parms.
 */
		while (*ptr && isspace (*ptr))
			++ptr;

	                ParseOption (option, *ptr ? ptr : NULL);
	}

	fclose (file);
}
////////////////////////////////////////////////////////////////////////////////
static void Usage (const char *error )
{
        int                     i;
        int                     max;
        struct OptionInfo*      info;

 	if (*error) fprintf(stderr,"\n%s\n",error);
 	fprintf(stderr,"Usage :\n");

        max = sizeof (optionTable) / sizeof (struct OptionInfo);
        for (i = 0, info = optionTable; i < max; i++, info++) {

                if (info->shortName)
                        fprintf (stderr, "-%-5s %s\n", info->shortName,
                                                        info->parmDescription);

                fprintf (stderr, "      %s\n\n", info->description);
        }

        exit (1);
}
////////////////////////////////////////////////////////////////////////////////
