/* Copyright (c) 2003 Andrew Stikheev. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * $Id: statistics.c,v 1.26 2004/09/07 09:44:41 sand Exp $
 */

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>
#include <errno.h>
#include <time.h>
#include <sys/time.h>
#include <string.h>
#include <arpa/inet.h>
#include <unistd.h>

#include "statistics.h"
#include "base.h"

int                     PutStatOut (statisticItem *, connectItem *, int);
int 			PutBlockOut(statisticItem *,  int);
static int              UpdateItem (statisticItem *);
static inline int	UpdateWeight(time_t, int, int);

extern systemParameters	*sys;

extern void		*hashTable;
extern int		fdLog;
//extern int		blockLog;
static  in_addr_t 	srcIP, dstIP;
static  u_short 	srcPort, dstPort;
static  int		type;
////////////////////////////////////////////////////////////////////////////////
int  HandlePacket(char *packet)
{
	static cluster		*cl;
	static statisticItem	*s,*best;
	static signed char	i;

	static struct ip	*ip;
	static struct tcphdr	*tcp;
	static struct udphdr	*udp;
//	static struct icmp	*icmp;

        ip = ( struct ip * ) packet;

	if ( ip->ip_p == IPPROTO_TCP ) {
   
                tcp = ( struct tcphdr *)(packet + sizeof ( struct ip ));

		if ( tcp->th_flags & ~TH_SYN ) return 0;

		type 	= 0;
		srcPort = tcp->th_sport;
		dstPort = tcp->th_dport;

        } else
        if ( ip->ip_p == IPPROTO_UDP ) {
		
		udp = ( struct udphdr *)(packet + sizeof( struct ip ));

		type 	= 1;
		srcPort = udp->uh_sport;
		dstPort = udp->uh_dport;

	} else
	if ( ip->ip_p == IPPROTO_ICMP ) {

//		icmp = ( struct icmp *)(packet + sizeof( struct ip));

		type 	= 2;
		srcPort = 0;
		dstPort = 0;

	} else return 0;

	srcIP = ip->ip_src.s_addr;
	dstIP = ip->ip_dst.s_addr;

	cl = (cluster *)hashTable + Hash( srcIP );
	// try to find first item with same ip or removed by scandctl item
	s = cl->item;
	best = NULL;
	for( i = 0 ; i < cl->total; i++,s++ ) { 
		if (  s->srcIP == srcIP ) {
			
			best = s;
			break;

		} else if ( !s->srcIP && best == NULL ) 
			best = s;
	}

	// if cluster is not full then we add new item


	if ( best == NULL && cl->total < sys->clusterSize )
		best = cl->item + cl->total++;

// cluster is full and need to add new item, so we find item with minimal weight

	if ( best == NULL ) {

		best = cl->item;
		s = best + 1;

		for( i = 1 ; i < sys->clusterSize ; i++,s++ ) {

		if ( best->weight[0] > s->weight[0] ) {

				best = s;
			}
		}
	}
	
	return UpdateItem ( best );

}
///////////////////////////////////////////////////////////////////////////////
int    UpdateItem( statisticItem *cur )
{
	static time_t		ctime;
	static time_t		lastTime;
	static int 		block;
	static int 		prevblock;
	static int 		prevblock_t;
	static int		i;
	static connectItem	*con,*c,*oldest;
	static int		InHash;

	block 		= 0; // not block packet
	prevblock 	= 0; // previus block status by default
	prevblock_t 	= 0; // previus block status of type by default

	time(&ctime);
	con = cur->con[type];

        if ( sys->blockIfAny ) {
		for ( i = 0 ; i < NUM_TYPES; i++ )
			if (cur->weight[i] >= sys->blkWeight[i]) {
				prevblock = 1;
				break;
			} 
		if ( cur->weight[type] >= sys->blkWeight[type]) prevblock_t = 1;
	}
	else {
		if ( cur->weight[type] >= sys->blkWeight[type] ) {
			prevblock   = 1;
			prevblock_t = 1;
		}
	}
	
	if ( cur->srcIP != srcIP ) {

		cur->srcIP  = srcIP;
		for( i = 0; i < NUM_TYPES; i++) {

			cur->weight[i]	= sys->minWeight[i];
			cur->total[i] 	= 0;
		}

		cur->total[type]	= 1;
		con->dstIP		= dstIP;
		con->dstPort		= dstPort;
		con->srcPort		= srcPort;
		con->time		= ctime;

	        if ( sys->blkFile[0] && sys->enableBlock )
                        PutBlockOut(cur, 0);

	}
	else {
		InHash	   = 0;
		c 	   = con;
		oldest	   = con;
		lastTime   = c->time;

		for ( i = 0 ; i <  cur->total[type] ; i++,c++ ) {

			if ( c->time > lastTime )  lastTime = c->time;
			if ( c->time < oldest->time ) oldest = c;

			if ( c->dstIP == dstIP && c->dstPort == dstPort ) {

				c->time = ctime;
				InHash  = 1;
				break;
			}
		}

		if ( !InHash ) {

			if (  cur->total[type] < sys->maxConnects[type] )
				con += cur->total[type]++;
			else  con = oldest;

			con->dstIP	= dstIP;
			con->dstPort	= dstPort;
			con->srcPort	= srcPort;
			con->time	= ctime;
		}

		cur->weight[type] += UpdateWeight (	ctime - lastTime, 
							type, 
							InHash );

		if ( 	  cur->weight[type] >= sys->maxWeight[type] ) 
			  cur->weight[type]  = sys->maxWeight[type];

		else if ( cur->weight[type]  < sys->minWeight[type] )
			  cur->weight[type]  = sys->minWeight[type];

		else if ( cur->weight[type] >= sys->blkWeight[type]  && 
			  !prevblock_t )
			  cur->weight[type]  = sys->maxWeight[type];

		if ( sys->blockIfAny ) {
			for ( i = 0 ; i < NUM_TYPES; i++ )
				if (cur->weight[i] >= sys->blkWeight[i]) {
					block = 1;
					break;
				}
		} else if ( cur->weight[type] >= sys->blkWeight[type] ) 
				block = 1;
	}

	if ( fdLog >= 0 ) PutStatOut(cur, con, sys->enableBlock ? block : 0 );
	if ( sys->blkFile[0] && sys->enableBlock && (block != prevblock) )
			PutBlockOut(cur, block);

	return block;
}
///////////////////////////////////////////////////////////////////////////////
int PutBlockOut(statisticItem *cur,  int block )
{
        static char             buf[256];       // i hope it's enough
        static char             size, stm[20];
        static time_t           tm;
        static struct in_addr   ip_src;
        static char             ipSrc[16];
	static int		fd = -1;
	static char		*openFile;

        ip_src.s_addr = cur->srcIP;

        time(&tm);

        strftime(stm, sizeof(stm)-1, "%Y-%m-%d %T", localtime(&tm));
        strcpy(ipSrc,(char *)inet_ntoa(ip_src));

        size=sprintf(buf,"%s\t%s\t%d\n", ipSrc,stm, block);

	openFile = sys->blTFile;

	if ( rename(sys->blkFile,sys->blTFile) == -1 ) {
		if ( errno == ENOENT )  
			openFile = sys->blkFile;
		 else 
			Quit("rename");
	}

	
	fd = open(	openFile, 
			O_NONBLOCK | O_CREAT | O_APPEND | O_WRONLY | O_FSYNC,
			S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH ); 

	if (fd < 0)  Quit("open");
	if ( write(fd, buf, size) != size ) Quit("write");
	close(fd);

	if ( openFile == sys->blTFile ) {
        	if ( rename( sys->blTFile, sys->blkFile ) == -1 ) 
                        Quit("rename");
	}

        return 0;
}

///////////////////////////////////////////////////////////////////////////////
int PutStatOut(statisticItem *cur, connectItem *con, int block )
{
	static char 		buf[256];  	// i hope it's enough
	static char 		size, stm[20];
	static time_t 		tm;
	static struct in_addr	ip_src,ip_dst;
	static char		ipSrc[16],ipDst[16];


	ip_src.s_addr = cur->srcIP;
	ip_dst.s_addr = con->dstIP;

        time(&tm);

	strftime(stm, sizeof(stm)-1, "%b %d %H:%M:%S", localtime(&tm));
	strcpy(ipSrc,(char *)inet_ntoa(ip_src));
	strcpy(ipDst,(char *)inet_ntoa(ip_dst));

	size=sprintf(buf,"%s :%d  %s:%d > %s:%d %d %s\n", stm, type,
			ipSrc, htons(con->srcPort), 
			ipDst, htons(con->dstPort),
			cur->weight[type],
			block ? "  blocked":  ""    );

        if (write(fdLog, buf, size) != size) {

                if (errno == EAGAIN || errno == ENOSPC || errno == EDQUOT) {

                        return 0;
                }
//                Quit("write log failed");
         }
	return 1;
}
///////////////////////////////////////////////////////////////////////////////
inline int UpdateWeight ( time_t delta, int typeOfPacket,int alreadyInHash )
{
  return  sys->factor[typeOfPacket][alreadyInHash] - delta;
}
///////////////////////////////////////////////////////////////////////////////
inline int Hash(in_addr_t ip)
{
        return ip % sys->numClusters;
}
///////////////////////////////////////////////////////////////////////////////
